# coding: utf-8

"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 0.1.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501

import warnings
from pydantic import validate_call, Field, StrictFloat, StrictStr, StrictInt
from typing import Any, Dict, List, Optional, Tuple, Union
from typing_extensions import Annotated

from pydantic import Field, StrictBool, StrictFloat, StrictInt, StrictStr
from typing import Any, List, Optional, Union
from typing_extensions import Annotated
from inductiva.client.models.available_simulator_versions import AvailableSimulatorVersions
from inductiva.client.models.bulk_simulator_operation import BulkSimulatorOperation
from inductiva.client.models.bulk_simulator_response import BulkSimulatorResponse
from inductiva.client.models.machine_type import MachineType
from inductiva.client.models.providers import Providers
from inductiva.client.models.simulator_create import SimulatorCreate
from inductiva.client.models.simulator_response import SimulatorResponse
from inductiva.client.models.simulator_update import SimulatorUpdate
from inductiva.client.models.simulator_with_machine_types import SimulatorWithMachineTypes

from inductiva.client.api_client import ApiClient, RequestSerialized
from inductiva.client.api_response import ApiResponse
from inductiva.client.rest import RESTResponseType


class SimulatorsApi:
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None) -> None:
        if api_client is None:
            api_client = ApiClient.get_default()
        self.api_client = api_client

    @validate_call
    def bulk_archive_simulators(
        self,
        bulk_simulator_operation: BulkSimulatorOperation,
        archive: Optional[StrictBool] = None,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> BulkSimulatorResponse:
        """Bulk Archive Simulators

        Bulk archive/unarchive simulators.

        :param bulk_simulator_operation: (required)
        :type bulk_simulator_operation: BulkSimulatorOperation
        :param archive:
        :type archive: bool
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._bulk_archive_simulators_serialize(
            bulk_simulator_operation=bulk_simulator_operation,
            archive=archive,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "BulkSimulatorResponse",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data

    @validate_call
    def bulk_archive_simulators_with_http_info(
        self,
        bulk_simulator_operation: BulkSimulatorOperation,
        archive: Optional[StrictBool] = None,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[BulkSimulatorResponse]:
        """Bulk Archive Simulators

        Bulk archive/unarchive simulators.

        :param bulk_simulator_operation: (required)
        :type bulk_simulator_operation: BulkSimulatorOperation
        :param archive:
        :type archive: bool
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._bulk_archive_simulators_serialize(
            bulk_simulator_operation=bulk_simulator_operation,
            archive=archive,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "BulkSimulatorResponse",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )

    @validate_call
    def bulk_archive_simulators_without_preload_content(
        self,
        bulk_simulator_operation: BulkSimulatorOperation,
        archive: Optional[StrictBool] = None,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Bulk Archive Simulators

        Bulk archive/unarchive simulators.

        :param bulk_simulator_operation: (required)
        :type bulk_simulator_operation: BulkSimulatorOperation
        :param archive:
        :type archive: bool
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._bulk_archive_simulators_serialize(
            bulk_simulator_operation=bulk_simulator_operation,
            archive=archive,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "BulkSimulatorResponse",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        return response_data.response

    def _bulk_archive_simulators_serialize(
        self,
        bulk_simulator_operation,
        archive,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:

        _host = None

        _collection_formats: Dict[str, str] = {}

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[str, Union[str, bytes, List[str], List[bytes],
                                List[Tuple[str, bytes]]]] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        # process the query parameters
        if archive is not None:

            _query_params.append(('archive', archive))

        # process the header parameters
        # process the form parameters
        # process the body parameter
        if bulk_simulator_operation is not None:
            _body_params = bulk_simulator_operation

        # set the HTTP header `Accept`
        if 'Accept' not in _header_params:
            _header_params['Accept'] = self.api_client.select_header_accept(
                ['application/json'])

        # set the HTTP header `Content-Type`
        if _content_type:
            _header_params['Content-Type'] = _content_type
        else:
            _default_content_type = (self.api_client.select_header_content_type(
                ['application/json']))
            if _default_content_type is not None:
                _header_params['Content-Type'] = _default_content_type

        # authentication setting
        _auth_settings: List[str] = ['APIKeyHeader']

        return self.api_client.param_serialize(
            method='POST',
            resource_path='/simulators/bulk/archive',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth)

    @validate_call
    def bulk_delete_simulators(
        self,
        bulk_simulator_operation: BulkSimulatorOperation,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> BulkSimulatorResponse:
        """Bulk Delete Simulators

        Bulk delete simulators.

        :param bulk_simulator_operation: (required)
        :type bulk_simulator_operation: BulkSimulatorOperation
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._bulk_delete_simulators_serialize(
            bulk_simulator_operation=bulk_simulator_operation,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "BulkSimulatorResponse",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data

    @validate_call
    def bulk_delete_simulators_with_http_info(
        self,
        bulk_simulator_operation: BulkSimulatorOperation,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[BulkSimulatorResponse]:
        """Bulk Delete Simulators

        Bulk delete simulators.

        :param bulk_simulator_operation: (required)
        :type bulk_simulator_operation: BulkSimulatorOperation
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._bulk_delete_simulators_serialize(
            bulk_simulator_operation=bulk_simulator_operation,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "BulkSimulatorResponse",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )

    @validate_call
    def bulk_delete_simulators_without_preload_content(
        self,
        bulk_simulator_operation: BulkSimulatorOperation,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Bulk Delete Simulators

        Bulk delete simulators.

        :param bulk_simulator_operation: (required)
        :type bulk_simulator_operation: BulkSimulatorOperation
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._bulk_delete_simulators_serialize(
            bulk_simulator_operation=bulk_simulator_operation,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "BulkSimulatorResponse",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        return response_data.response

    def _bulk_delete_simulators_serialize(
        self,
        bulk_simulator_operation,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:

        _host = None

        _collection_formats: Dict[str, str] = {}

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[str, Union[str, bytes, List[str], List[bytes],
                                List[Tuple[str, bytes]]]] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        # process the query parameters
        # process the header parameters
        # process the form parameters
        # process the body parameter
        if bulk_simulator_operation is not None:
            _body_params = bulk_simulator_operation

        # set the HTTP header `Accept`
        if 'Accept' not in _header_params:
            _header_params['Accept'] = self.api_client.select_header_accept(
                ['application/json'])

        # set the HTTP header `Content-Type`
        if _content_type:
            _header_params['Content-Type'] = _content_type
        else:
            _default_content_type = (self.api_client.select_header_content_type(
                ['application/json']))
            if _default_content_type is not None:
                _header_params['Content-Type'] = _default_content_type

        # authentication setting
        _auth_settings: List[str] = ['APIKeyHeader']

        return self.api_client.param_serialize(
            method='POST',
            resource_path='/simulators/bulk/delete',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth)

    @validate_call
    def create_simulator(
        self,
        simulator_create: SimulatorCreate,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> SimulatorResponse:
        """Create Simulator

        Create a new simulator.

        :param simulator_create: (required)
        :type simulator_create: SimulatorCreate
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._create_simulator_serialize(
            simulator_create=simulator_create,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '201': "SimulatorResponse",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data

    @validate_call
    def create_simulator_with_http_info(
        self,
        simulator_create: SimulatorCreate,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[SimulatorResponse]:
        """Create Simulator

        Create a new simulator.

        :param simulator_create: (required)
        :type simulator_create: SimulatorCreate
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._create_simulator_serialize(
            simulator_create=simulator_create,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '201': "SimulatorResponse",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )

    @validate_call
    def create_simulator_without_preload_content(
        self,
        simulator_create: SimulatorCreate,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Create Simulator

        Create a new simulator.

        :param simulator_create: (required)
        :type simulator_create: SimulatorCreate
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._create_simulator_serialize(
            simulator_create=simulator_create,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '201': "SimulatorResponse",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        return response_data.response

    def _create_simulator_serialize(
        self,
        simulator_create,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:

        _host = None

        _collection_formats: Dict[str, str] = {}

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[str, Union[str, bytes, List[str], List[bytes],
                                List[Tuple[str, bytes]]]] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        # process the query parameters
        # process the header parameters
        # process the form parameters
        # process the body parameter
        if simulator_create is not None:
            _body_params = simulator_create

        # set the HTTP header `Accept`
        if 'Accept' not in _header_params:
            _header_params['Accept'] = self.api_client.select_header_accept(
                ['application/json'])

        # set the HTTP header `Content-Type`
        if _content_type:
            _header_params['Content-Type'] = _content_type
        else:
            _default_content_type = (self.api_client.select_header_content_type(
                ['application/json']))
            if _default_content_type is not None:
                _header_params['Content-Type'] = _default_content_type

        # authentication setting
        _auth_settings: List[str] = ['APIKeyHeader']

        return self.api_client.param_serialize(
            method='POST',
            resource_path='/simulators/',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth)

    @validate_call
    def delete_simulator(
        self,
        simulator_id: StrictStr,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> None:
        """Delete Simulator

        Delete a simulator.

        :param simulator_id: (required)
        :type simulator_id: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._delete_simulator_serialize(simulator_id=simulator_id,
                                                  _request_auth=_request_auth,
                                                  _content_type=_content_type,
                                                  _headers=_headers,
                                                  _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '204': None,
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data

    @validate_call
    def delete_simulator_with_http_info(
        self,
        simulator_id: StrictStr,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[None]:
        """Delete Simulator

        Delete a simulator.

        :param simulator_id: (required)
        :type simulator_id: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._delete_simulator_serialize(simulator_id=simulator_id,
                                                  _request_auth=_request_auth,
                                                  _content_type=_content_type,
                                                  _headers=_headers,
                                                  _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '204': None,
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )

    @validate_call
    def delete_simulator_without_preload_content(
        self,
        simulator_id: StrictStr,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Delete Simulator

        Delete a simulator.

        :param simulator_id: (required)
        :type simulator_id: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._delete_simulator_serialize(simulator_id=simulator_id,
                                                  _request_auth=_request_auth,
                                                  _content_type=_content_type,
                                                  _headers=_headers,
                                                  _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '204': None,
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        return response_data.response

    def _delete_simulator_serialize(
        self,
        simulator_id,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:

        _host = None

        _collection_formats: Dict[str, str] = {}

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[str, Union[str, bytes, List[str], List[bytes],
                                List[Tuple[str, bytes]]]] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        if simulator_id is not None:
            _path_params['simulator_id'] = simulator_id
        # process the query parameters
        # process the header parameters
        # process the form parameters
        # process the body parameter

        # set the HTTP header `Accept`
        if 'Accept' not in _header_params:
            _header_params['Accept'] = self.api_client.select_header_accept(
                ['application/json'])

        # authentication setting
        _auth_settings: List[str] = ['APIKeyHeader']

        return self.api_client.param_serialize(
            method='DELETE',
            resource_path='/simulators/{simulator_id}',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth)

    @validate_call
    def get_simulator_with_machine_types(
        self,
        simulator_name: StrictStr,
        machine_families: Optional[List[StrictStr]] = None,
        machine_configs: Optional[List[StrictStr]] = None,
        vcpus_range: Optional[Annotated[
            List[StrictInt], Field(min_length=2, max_length=2)]] = None,
        memory_range: Optional[Annotated[
            List[StrictInt], Field(min_length=2, max_length=2)]] = None,
        price_range: Optional[Annotated[
            List[Union[StrictFloat, StrictInt]],
            Field(min_length=2, max_length=2)]] = None,
        spot: Optional[Any] = None,
        gpus_range: Optional[Annotated[
            List[StrictInt], Field(min_length=2, max_length=2)]] = None,
        gpu_names: Optional[List[StrictStr]] = None,
        zones: Optional[List[StrictStr]] = None,
        provider_id: Optional[Providers] = None,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> SimulatorWithMachineTypes:
        """Get Simulator With Machine Types

        Get an available machine type for a particular simulator.

        :param simulator_name: (required)
        :type simulator_name: str
        :param machine_families:
        :type machine_families: List[str]
        :param machine_configs:
        :type machine_configs: List[str]
        :param vcpus_range:
        :type vcpus_range: List[int]
        :param memory_range:
        :type memory_range: List[int]
        :param price_range:
        :type price_range: List[float]
        :param spot:
        :type spot: Spot1
        :param gpus_range:
        :type gpus_range: List[int]
        :param gpu_names:
        :type gpu_names: List[str]
        :param zones:
        :type zones: List[str]
        :param provider_id:
        :type provider_id: Providers
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._get_simulator_with_machine_types_serialize(
            simulator_name=simulator_name,
            machine_families=machine_families,
            machine_configs=machine_configs,
            vcpus_range=vcpus_range,
            memory_range=memory_range,
            price_range=price_range,
            spot=spot,
            gpus_range=gpus_range,
            gpu_names=gpu_names,
            zones=zones,
            provider_id=provider_id,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "SimulatorWithMachineTypes",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data

    @validate_call
    def get_simulator_with_machine_types_with_http_info(
        self,
        simulator_name: StrictStr,
        machine_families: Optional[List[StrictStr]] = None,
        machine_configs: Optional[List[StrictStr]] = None,
        vcpus_range: Optional[Annotated[
            List[StrictInt], Field(min_length=2, max_length=2)]] = None,
        memory_range: Optional[Annotated[
            List[StrictInt], Field(min_length=2, max_length=2)]] = None,
        price_range: Optional[Annotated[
            List[Union[StrictFloat, StrictInt]],
            Field(min_length=2, max_length=2)]] = None,
        spot: Optional[Any] = None,
        gpus_range: Optional[Annotated[
            List[StrictInt], Field(min_length=2, max_length=2)]] = None,
        gpu_names: Optional[List[StrictStr]] = None,
        zones: Optional[List[StrictStr]] = None,
        provider_id: Optional[Providers] = None,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[SimulatorWithMachineTypes]:
        """Get Simulator With Machine Types

        Get an available machine type for a particular simulator.

        :param simulator_name: (required)
        :type simulator_name: str
        :param machine_families:
        :type machine_families: List[str]
        :param machine_configs:
        :type machine_configs: List[str]
        :param vcpus_range:
        :type vcpus_range: List[int]
        :param memory_range:
        :type memory_range: List[int]
        :param price_range:
        :type price_range: List[float]
        :param spot:
        :type spot: Spot1
        :param gpus_range:
        :type gpus_range: List[int]
        :param gpu_names:
        :type gpu_names: List[str]
        :param zones:
        :type zones: List[str]
        :param provider_id:
        :type provider_id: Providers
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._get_simulator_with_machine_types_serialize(
            simulator_name=simulator_name,
            machine_families=machine_families,
            machine_configs=machine_configs,
            vcpus_range=vcpus_range,
            memory_range=memory_range,
            price_range=price_range,
            spot=spot,
            gpus_range=gpus_range,
            gpu_names=gpu_names,
            zones=zones,
            provider_id=provider_id,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "SimulatorWithMachineTypes",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )

    @validate_call
    def get_simulator_with_machine_types_without_preload_content(
        self,
        simulator_name: StrictStr,
        machine_families: Optional[List[StrictStr]] = None,
        machine_configs: Optional[List[StrictStr]] = None,
        vcpus_range: Optional[Annotated[
            List[StrictInt], Field(min_length=2, max_length=2)]] = None,
        memory_range: Optional[Annotated[
            List[StrictInt], Field(min_length=2, max_length=2)]] = None,
        price_range: Optional[Annotated[
            List[Union[StrictFloat, StrictInt]],
            Field(min_length=2, max_length=2)]] = None,
        spot: Optional[Any] = None,
        gpus_range: Optional[Annotated[
            List[StrictInt], Field(min_length=2, max_length=2)]] = None,
        gpu_names: Optional[List[StrictStr]] = None,
        zones: Optional[List[StrictStr]] = None,
        provider_id: Optional[Providers] = None,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Get Simulator With Machine Types

        Get an available machine type for a particular simulator.

        :param simulator_name: (required)
        :type simulator_name: str
        :param machine_families:
        :type machine_families: List[str]
        :param machine_configs:
        :type machine_configs: List[str]
        :param vcpus_range:
        :type vcpus_range: List[int]
        :param memory_range:
        :type memory_range: List[int]
        :param price_range:
        :type price_range: List[float]
        :param spot:
        :type spot: Spot1
        :param gpus_range:
        :type gpus_range: List[int]
        :param gpu_names:
        :type gpu_names: List[str]
        :param zones:
        :type zones: List[str]
        :param provider_id:
        :type provider_id: Providers
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._get_simulator_with_machine_types_serialize(
            simulator_name=simulator_name,
            machine_families=machine_families,
            machine_configs=machine_configs,
            vcpus_range=vcpus_range,
            memory_range=memory_range,
            price_range=price_range,
            spot=spot,
            gpus_range=gpus_range,
            gpu_names=gpu_names,
            zones=zones,
            provider_id=provider_id,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "SimulatorWithMachineTypes",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        return response_data.response

    def _get_simulator_with_machine_types_serialize(
        self,
        simulator_name,
        machine_families,
        machine_configs,
        vcpus_range,
        memory_range,
        price_range,
        spot,
        gpus_range,
        gpu_names,
        zones,
        provider_id,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:

        _host = None

        _collection_formats: Dict[str, str] = {
            'machine_families': 'multi',
            'machine_configs': 'multi',
            'vcpus_range': 'multi',
            'memory_range': 'multi',
            'price_range': 'multi',
            'gpus_range': 'multi',
            'gpu_names': 'multi',
            'zones': 'multi',
        }

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[str, Union[str, bytes, List[str], List[bytes],
                                List[Tuple[str, bytes]]]] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        if simulator_name is not None:
            _path_params['simulator_name'] = simulator_name
        # process the query parameters
        if machine_families is not None:

            _query_params.append(('machine_families', machine_families))

        if machine_configs is not None:

            _query_params.append(('machine_configs', machine_configs))

        if vcpus_range is not None:

            _query_params.append(('vcpus_range', vcpus_range))

        if memory_range is not None:

            _query_params.append(('memory_range', memory_range))

        if price_range is not None:

            _query_params.append(('price_range', price_range))

        if spot is not None:

            _query_params.append(('spot', spot))

        if gpus_range is not None:

            _query_params.append(('gpus_range', gpus_range))

        if gpu_names is not None:

            _query_params.append(('gpu_names', gpu_names))

        if zones is not None:

            _query_params.append(('zones', zones))

        if provider_id is not None:

            _query_params.append(('provider_id', provider_id.value))

        # process the header parameters
        # process the form parameters
        # process the body parameter

        # set the HTTP header `Accept`
        if 'Accept' not in _header_params:
            _header_params['Accept'] = self.api_client.select_header_accept(
                ['application/json'])

        # authentication setting
        _auth_settings: List[str] = []

        return self.api_client.param_serialize(
            method='GET',
            resource_path='/simulators/{simulator_name}/machine-types',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth)

    @validate_call
    def list_all_machine_types(
        self,
        machine_families: Optional[List[StrictStr]] = None,
        machine_configs: Optional[List[StrictStr]] = None,
        vcpus_range: Optional[Annotated[
            List[StrictInt], Field(min_length=2, max_length=2)]] = None,
        memory_range: Optional[Annotated[
            List[StrictInt], Field(min_length=2, max_length=2)]] = None,
        price_range: Optional[Annotated[
            List[Union[StrictFloat, StrictInt]],
            Field(min_length=2, max_length=2)]] = None,
        spot: Optional[Any] = None,
        gpus_range: Optional[Annotated[
            List[StrictInt], Field(min_length=2, max_length=2)]] = None,
        gpu_names: Optional[List[StrictStr]] = None,
        zones: Optional[List[StrictStr]] = None,
        provider_id: Optional[Providers] = None,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> List[MachineType]:
        """List All Machine Types

        List available machine types for the given provider and zone.

        :param machine_families:
        :type machine_families: List[str]
        :param machine_configs:
        :type machine_configs: List[str]
        :param vcpus_range:
        :type vcpus_range: List[int]
        :param memory_range:
        :type memory_range: List[int]
        :param price_range:
        :type price_range: List[float]
        :param spot:
        :type spot: Spot1
        :param gpus_range:
        :type gpus_range: List[int]
        :param gpu_names:
        :type gpu_names: List[str]
        :param zones:
        :type zones: List[str]
        :param provider_id:
        :type provider_id: Providers
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._list_all_machine_types_serialize(
            machine_families=machine_families,
            machine_configs=machine_configs,
            vcpus_range=vcpus_range,
            memory_range=memory_range,
            price_range=price_range,
            spot=spot,
            gpus_range=gpus_range,
            gpu_names=gpu_names,
            zones=zones,
            provider_id=provider_id,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "List[MachineType]",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data

    @validate_call
    def list_all_machine_types_with_http_info(
        self,
        machine_families: Optional[List[StrictStr]] = None,
        machine_configs: Optional[List[StrictStr]] = None,
        vcpus_range: Optional[Annotated[
            List[StrictInt], Field(min_length=2, max_length=2)]] = None,
        memory_range: Optional[Annotated[
            List[StrictInt], Field(min_length=2, max_length=2)]] = None,
        price_range: Optional[Annotated[
            List[Union[StrictFloat, StrictInt]],
            Field(min_length=2, max_length=2)]] = None,
        spot: Optional[Any] = None,
        gpus_range: Optional[Annotated[
            List[StrictInt], Field(min_length=2, max_length=2)]] = None,
        gpu_names: Optional[List[StrictStr]] = None,
        zones: Optional[List[StrictStr]] = None,
        provider_id: Optional[Providers] = None,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[List[MachineType]]:
        """List All Machine Types

        List available machine types for the given provider and zone.

        :param machine_families:
        :type machine_families: List[str]
        :param machine_configs:
        :type machine_configs: List[str]
        :param vcpus_range:
        :type vcpus_range: List[int]
        :param memory_range:
        :type memory_range: List[int]
        :param price_range:
        :type price_range: List[float]
        :param spot:
        :type spot: Spot1
        :param gpus_range:
        :type gpus_range: List[int]
        :param gpu_names:
        :type gpu_names: List[str]
        :param zones:
        :type zones: List[str]
        :param provider_id:
        :type provider_id: Providers
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._list_all_machine_types_serialize(
            machine_families=machine_families,
            machine_configs=machine_configs,
            vcpus_range=vcpus_range,
            memory_range=memory_range,
            price_range=price_range,
            spot=spot,
            gpus_range=gpus_range,
            gpu_names=gpu_names,
            zones=zones,
            provider_id=provider_id,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "List[MachineType]",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )

    @validate_call
    def list_all_machine_types_without_preload_content(
        self,
        machine_families: Optional[List[StrictStr]] = None,
        machine_configs: Optional[List[StrictStr]] = None,
        vcpus_range: Optional[Annotated[
            List[StrictInt], Field(min_length=2, max_length=2)]] = None,
        memory_range: Optional[Annotated[
            List[StrictInt], Field(min_length=2, max_length=2)]] = None,
        price_range: Optional[Annotated[
            List[Union[StrictFloat, StrictInt]],
            Field(min_length=2, max_length=2)]] = None,
        spot: Optional[Any] = None,
        gpus_range: Optional[Annotated[
            List[StrictInt], Field(min_length=2, max_length=2)]] = None,
        gpu_names: Optional[List[StrictStr]] = None,
        zones: Optional[List[StrictStr]] = None,
        provider_id: Optional[Providers] = None,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """List All Machine Types

        List available machine types for the given provider and zone.

        :param machine_families:
        :type machine_families: List[str]
        :param machine_configs:
        :type machine_configs: List[str]
        :param vcpus_range:
        :type vcpus_range: List[int]
        :param memory_range:
        :type memory_range: List[int]
        :param price_range:
        :type price_range: List[float]
        :param spot:
        :type spot: Spot1
        :param gpus_range:
        :type gpus_range: List[int]
        :param gpu_names:
        :type gpu_names: List[str]
        :param zones:
        :type zones: List[str]
        :param provider_id:
        :type provider_id: Providers
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._list_all_machine_types_serialize(
            machine_families=machine_families,
            machine_configs=machine_configs,
            vcpus_range=vcpus_range,
            memory_range=memory_range,
            price_range=price_range,
            spot=spot,
            gpus_range=gpus_range,
            gpu_names=gpu_names,
            zones=zones,
            provider_id=provider_id,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "List[MachineType]",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        return response_data.response

    def _list_all_machine_types_serialize(
        self,
        machine_families,
        machine_configs,
        vcpus_range,
        memory_range,
        price_range,
        spot,
        gpus_range,
        gpu_names,
        zones,
        provider_id,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:

        _host = None

        _collection_formats: Dict[str, str] = {
            'machine_families': 'multi',
            'machine_configs': 'multi',
            'vcpus_range': 'multi',
            'memory_range': 'multi',
            'price_range': 'multi',
            'gpus_range': 'multi',
            'gpu_names': 'multi',
            'zones': 'multi',
        }

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[str, Union[str, bytes, List[str], List[bytes],
                                List[Tuple[str, bytes]]]] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        # process the query parameters
        if machine_families is not None:

            _query_params.append(('machine_families', machine_families))

        if machine_configs is not None:

            _query_params.append(('machine_configs', machine_configs))

        if vcpus_range is not None:

            _query_params.append(('vcpus_range', vcpus_range))

        if memory_range is not None:

            _query_params.append(('memory_range', memory_range))

        if price_range is not None:

            _query_params.append(('price_range', price_range))

        if spot is not None:

            _query_params.append(('spot', spot))

        if gpus_range is not None:

            _query_params.append(('gpus_range', gpus_range))

        if gpu_names is not None:

            _query_params.append(('gpu_names', gpu_names))

        if zones is not None:

            _query_params.append(('zones', zones))

        if provider_id is not None:

            _query_params.append(('provider_id', provider_id.value))

        # process the header parameters
        # process the form parameters
        # process the body parameter

        # set the HTTP header `Accept`
        if 'Accept' not in _header_params:
            _header_params['Accept'] = self.api_client.select_header_accept(
                ['application/json'])

        # authentication setting
        _auth_settings: List[str] = []

        return self.api_client.param_serialize(
            method='GET',
            resource_path='/simulators/machine-types',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth)

    @validate_call
    def list_available_images(
        self,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> AvailableSimulatorVersions:
        """List Available Images

        List available simulator versions with fallback mechanism.  First tries to get images from DockerHub API, if that fails, falls back to the database.  Returns:     A dictionary with the available simulator versions, grouped by branch     (development/production) and simulator name. {     {'development':         {'amr-wind': ['1.4.0'],           ...          'dualsphysics': ['5.2.1']         },     'production':         {'amr-wind': ['1.3.0', '1.4.0'],           ...          'dualsphysics': ['5.2.1']         }     } }

        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._list_available_images_serialize(
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "AvailableSimulatorVersions",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data

    @validate_call
    def list_available_images_with_http_info(
        self,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[AvailableSimulatorVersions]:
        """List Available Images

        List available simulator versions with fallback mechanism.  First tries to get images from DockerHub API, if that fails, falls back to the database.  Returns:     A dictionary with the available simulator versions, grouped by branch     (development/production) and simulator name. {     {'development':         {'amr-wind': ['1.4.0'],           ...          'dualsphysics': ['5.2.1']         },     'production':         {'amr-wind': ['1.3.0', '1.4.0'],           ...          'dualsphysics': ['5.2.1']         }     } }

        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._list_available_images_serialize(
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "AvailableSimulatorVersions",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )

    @validate_call
    def list_available_images_without_preload_content(
        self,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """List Available Images

        List available simulator versions with fallback mechanism.  First tries to get images from DockerHub API, if that fails, falls back to the database.  Returns:     A dictionary with the available simulator versions, grouped by branch     (development/production) and simulator name. {     {'development':         {'amr-wind': ['1.4.0'],           ...          'dualsphysics': ['5.2.1']         },     'production':         {'amr-wind': ['1.3.0', '1.4.0'],           ...          'dualsphysics': ['5.2.1']         }     } }

        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._list_available_images_serialize(
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "AvailableSimulatorVersions",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        return response_data.response

    def _list_available_images_serialize(
        self,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:

        _host = None

        _collection_formats: Dict[str, str] = {}

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[str, Union[str, bytes, List[str], List[bytes],
                                List[Tuple[str, bytes]]]] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        # process the query parameters
        # process the header parameters
        # process the form parameters
        # process the body parameter

        # set the HTTP header `Accept`
        if 'Accept' not in _header_params:
            _header_params['Accept'] = self.api_client.select_header_accept(
                ['application/json'])

        # authentication setting
        _auth_settings: List[str] = []

        return self.api_client.param_serialize(
            method='GET',
            resource_path='/simulators/available-images',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth)

    @validate_call
    def list_simulators(
        self,
        name: Optional[StrictStr] = None,
        include_archived: Optional[StrictBool] = None,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> List[SimulatorResponse]:
        """List Simulators

        List all simulators.

        :param name:
        :type name: str
        :param include_archived:
        :type include_archived: bool
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._list_simulators_serialize(
            name=name,
            include_archived=include_archived,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "List[SimulatorResponse]",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data

    @validate_call
    def list_simulators_with_http_info(
        self,
        name: Optional[StrictStr] = None,
        include_archived: Optional[StrictBool] = None,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[List[SimulatorResponse]]:
        """List Simulators

        List all simulators.

        :param name:
        :type name: str
        :param include_archived:
        :type include_archived: bool
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._list_simulators_serialize(
            name=name,
            include_archived=include_archived,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "List[SimulatorResponse]",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )

    @validate_call
    def list_simulators_without_preload_content(
        self,
        name: Optional[StrictStr] = None,
        include_archived: Optional[StrictBool] = None,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """List Simulators

        List all simulators.

        :param name:
        :type name: str
        :param include_archived:
        :type include_archived: bool
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._list_simulators_serialize(
            name=name,
            include_archived=include_archived,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "List[SimulatorResponse]",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        return response_data.response

    def _list_simulators_serialize(
        self,
        name,
        include_archived,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:

        _host = None

        _collection_formats: Dict[str, str] = {}

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[str, Union[str, bytes, List[str], List[bytes],
                                List[Tuple[str, bytes]]]] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        # process the query parameters
        if name is not None:

            _query_params.append(('name', name))

        if include_archived is not None:

            _query_params.append(('include_archived', include_archived))

        # process the header parameters
        # process the form parameters
        # process the body parameter

        # set the HTTP header `Accept`
        if 'Accept' not in _header_params:
            _header_params['Accept'] = self.api_client.select_header_accept(
                ['application/json'])

        # authentication setting
        _auth_settings: List[str] = []

        return self.api_client.param_serialize(
            method='GET',
            resource_path='/simulators/',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth)

    @validate_call
    def update_simulator(
        self,
        simulator_id: StrictStr,
        simulator_update: SimulatorUpdate,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> SimulatorResponse:
        """Update Simulator

        Update a simulator.

        :param simulator_id: (required)
        :type simulator_id: str
        :param simulator_update: (required)
        :type simulator_update: SimulatorUpdate
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._update_simulator_serialize(
            simulator_id=simulator_id,
            simulator_update=simulator_update,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "SimulatorResponse",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data

    @validate_call
    def update_simulator_with_http_info(
        self,
        simulator_id: StrictStr,
        simulator_update: SimulatorUpdate,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[SimulatorResponse]:
        """Update Simulator

        Update a simulator.

        :param simulator_id: (required)
        :type simulator_id: str
        :param simulator_update: (required)
        :type simulator_update: SimulatorUpdate
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._update_simulator_serialize(
            simulator_id=simulator_id,
            simulator_update=simulator_update,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "SimulatorResponse",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )

    @validate_call
    def update_simulator_without_preload_content(
        self,
        simulator_id: StrictStr,
        simulator_update: SimulatorUpdate,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Update Simulator

        Update a simulator.

        :param simulator_id: (required)
        :type simulator_id: str
        :param simulator_update: (required)
        :type simulator_update: SimulatorUpdate
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._update_simulator_serialize(
            simulator_id=simulator_id,
            simulator_update=simulator_update,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "SimulatorResponse",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        return response_data.response

    def _update_simulator_serialize(
        self,
        simulator_id,
        simulator_update,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:

        _host = None

        _collection_formats: Dict[str, str] = {}

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[str, Union[str, bytes, List[str], List[bytes],
                                List[Tuple[str, bytes]]]] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        if simulator_id is not None:
            _path_params['simulator_id'] = simulator_id
        # process the query parameters
        # process the header parameters
        # process the form parameters
        # process the body parameter
        if simulator_update is not None:
            _body_params = simulator_update

        # set the HTTP header `Accept`
        if 'Accept' not in _header_params:
            _header_params['Accept'] = self.api_client.select_header_accept(
                ['application/json'])

        # set the HTTP header `Content-Type`
        if _content_type:
            _header_params['Content-Type'] = _content_type
        else:
            _default_content_type = (self.api_client.select_header_content_type(
                ['application/json']))
            if _default_content_type is not None:
                _header_params['Content-Type'] = _default_content_type

        # authentication setting
        _auth_settings: List[str] = ['APIKeyHeader']

        return self.api_client.param_serialize(
            method='PUT',
            resource_path='/simulators/{simulator_id}',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth)



"""Methods to interact with the user info on the API."""
import argparse

from inductiva._cli.cmd_auth.login import login as login_cmd
from inductiva._cli.cmd_auth.logout import logout as logout_cmd


def login(private: bool = False) -> None:
    """Logs the user in to the Inductiva platform.

    This function handles user login by prompting the user to log in via the
      command line.

    Args:
        private:  If True, no api_key will be printed to the console.
    """

    args = argparse.Namespace(private=private)
    login_cmd(args)


def logout():
    """Logs the user out by removing the stored API key."""
    logout_cmd(None)



"""Methods to interact with the user storage resources."""
import json
import datetime
import itertools
import logging
import math
import os
import pathlib
import threading
import time
import urllib
import zlib
from dataclasses import dataclass
from enum import Enum
from typing import List, Literal, Optional, Tuple

import concurrent.futures
import tqdm

import inductiva
from inductiva import constants
from inductiva import utils
from inductiva.api import methods
import inductiva.client
from inductiva.client import exceptions, models
from inductiva.utils import format_utils
from inductiva.utils import data

MB = 1024 * 1024
_boto3_imported = True
try:
    import boto3
    from botocore.config import Config
    logging.getLogger("botocore").setLevel(logging.WARNING)
except ImportError:
    _boto3_imported = False


def _print_storage_size_and_cost() -> int:
    """ Print the storage size and cost.

    return the storage size in bytes.
    """
    api = inductiva.client.StorageApi(inductiva.api.get_client())
    storage_total_size_bytes = api.get_storage_size()
    estimated_storage_cost = api.get_storage_monthly_cost(
    ).estimated_monthly_cost

    estimated_storage_cost = format_utils.currency_formatter(
        estimated_storage_cost)
    storage_total_size = format_utils.bytes_formatter(storage_total_size_bytes)

    print("Total storage size used:")
    print(f"\tVolume: {storage_total_size}")
    print(f"\tCost: {estimated_storage_cost}/month")
    print("")

    return storage_total_size_bytes


def get_space_used():
    """Returns the occupied storage size in GB."""
    storage_total_size_bytes = _print_storage_size_and_cost()
    #Return float instead of a string
    storage_used = round(float(storage_total_size_bytes) / (1024**3), 3)
    return storage_used


def listdir(
    path="/",
    max_results: Optional[int] = 10,
    order_by: Literal["size", "creation_time"] = "creation_time",
    sort_order: Literal["asc", "desc"] = "desc",
    print_results: bool = True,
):
    """List and display the contents of the user's storage.
    Args:
        path (str): Storage directory to list. Default is root.
        max_results (int): The maximum number of results to return. If not set,
            all entries are returned.
        order_by (str): The field to sort the contents by.
        sort_order (str): Whether to sort the contents in ascending or
        descending order.
    Returns:
        list of dict: A list of dictionaries containing information about
        the size, the name and the creation time of each content that can
        easily be converted to a dataframe.

    This function prints a table with the storage content information:
        Name            Size            Creation Time
        1234            5.68 MiB        29 Sep, 14:12:00
        12345           374.85 KiB      29 Sep, 14:13:10
        1234567         97.59 KiB       29 Sep, 14:13:24
        123             0 B             29 Sep, 14:13:29
        123456          0 B             29 Sep, 14:13:18
    You can use this information to delete the contents you don't need
    anymore and further inspect task outputs and logs using the Task
    class.
    """

    api = inductiva.client.StorageApi(inductiva.api.get_client())

    # This is valid for single files and directories
    if len(path.split("/")) < 2:
        path += "/"

    contents = api.list_storage_contents(path=path,
                                         sort_by=order_by,
                                         order=sort_order,
                                         max_results=max_results)
    all_contents = []
    for content_name, info in contents.items():
        size = info.size_bytes
        creation_time = info.creation_time
        all_contents.append({
            "content_name": content_name,
            "size": round(float(size), 3),
            "creation_time": creation_time
        })
    if print_results:
        print(_print_contents_table(all_contents))

        _print_storage_size_and_cost()

        print(
            f"Listed {len(all_contents)} folder(s). Ordered by {order_by}.\n"
            "Use --max-results/-m to control the number of results displayed.")
    return all_contents


def _print_contents_table(contents):
    columns = ["Name", "Size", "Creation Time"]
    rows = []

    for content in contents:
        row = [
            content["content_name"], content["size"], content["creation_time"]
        ]
        rows.append(row)

    formatters = {
        "Creation Time": [format_utils.datetime_formatter],
        "Size": [format_utils.bytes_formatter],
    }

    emph_formatter = format_utils.get_ansi_formatter()
    header_formatters = [
        lambda x: emph_formatter(x.upper(), format_utils.Emphasis.BOLD)
    ]

    return format_utils.get_tabular_str(
        rows,
        columns,
        formatters=formatters,
        header_formatters=header_formatters,
    )


def get_signed_urls(
    paths: List[str],
    operation: Literal["upload", "download"],
) -> List[str]:
    api_instance = inductiva.client.StorageApi(inductiva.api.get_client())
    resp = api_instance.get_signed_urls_without_preload_content(
        paths=paths,
        operation=models.OperationType(operation),
    )

    return json.loads(resp.data)


@dataclass
class ZipFileInfo:
    """Represents information about a file within a ZIP archive."""
    name: str
    size: Optional[int]
    compressed_size: Optional[int]
    range_start: Optional[int]
    creation_time: Optional[datetime.datetime]
    compress_type: Optional[int]


@dataclass
class ZipArchiveInfo:
    """Represents the total ZIP size and file contents of a ZIP archive."""
    size: int
    files: List[ZipFileInfo]


@dataclass
class ZipFileRange:
    """Represents the byte range of a file within a ZIP archive."""
    range_start: int
    range_end: int


def get_zip_contents(
    path: str,
    zip_relative_path: str = "",
    recursive: bool = False,
) -> ZipArchiveInfo:
    """
    Retrieve the contents of a ZIP archive from a given path.

    Args:
        path (str): The full path to the ZIP archive.
        zip_relative_path (str, optional): A relative path inside the ZIP
            archive to filter the contents. Defaults to an empty string,
            which lists all files within the archive.
        recursive (bool, optional): If True, list contents recursively within
            the specified `zip_relative_path`. If False, list only top-level
            files and directories within the specified `zip_relative_path`.
            Defaults to False.

    Returns:
        ZipArchiveInfo: An object containing the total size of the ZIP archive
            and a list of `ZipFileInfo` objects representing the files
            within the specified ZIP archive.
    """
    api_instance = inductiva.client.StorageApi(inductiva.api.get_client())
    archive_info = api_instance.get_zip_contents(
        path=path,
        zip_relative_path=zip_relative_path,
        recursive=str(recursive).lower())
    files = [
        ZipFileInfo(
            name=file.name,
            size=file.size,
            compressed_size=file.compressed_size,
            range_start=file.range_start,
            creation_time=file.creation_time,
            compress_type=file.compress_type,
        ) for file in archive_info.contents
    ]
    return ZipArchiveInfo(size=archive_info.size, files=files)


def get_file_range(
    path: str,
    zip_relative_path: str = "",
    filename: str = "",
) -> ZipFileRange:
    """
    Retrieve the byte range (start and end) of the compressed data for a
    specific file inside a ZIP archive.

    Args:
        path (str): The full path to the ZIP archive.
        zip_relative_path (str, optional): The relative path inside the ZIP.
        filename (str): The name of the file inside the ZIP to get the range.

    Returns:
        ZipFileRange: The start and end byte offsets of the file
    """
    api_instance = inductiva.client.StorageApi(inductiva.api.get_client())
    response = api_instance.get_zip_file_range(
        filename=filename,
        zip_relative_path=zip_relative_path,
        path=path,
    )
    return ZipFileRange(
        range_start=response.range_start,
        range_end=response.range_end,
    )


def upload_from_url(
    url: str,
    remote_dir: str,
    file_name: Optional[str] = None,
):
    """
    Upload a file from a given URL to a specified remote directory.

    If no file name is provided, the function extracts the name from the URL.

    Args:
        url (str): The URL of the file to be uploaded.
        remote_dir (str): The path to the remote directory where the file will
            be stored.
        file_name (str, optional): The name to save the uploaded file as.
            If not specified, the name will be extracted from the URL.
    """
    api_instance = inductiva.client.StorageApi(inductiva.api.get_client())

    # append filename extracted from url to remote_path
    if not file_name:
        parsed_url = urllib.parse.urlparse(url)
        file_name = urllib.parse.unquote(parsed_url.path.split("/")[-1])

    remote_path = os.path.join(remote_dir, file_name)

    api_instance.upload_from_url(url=url, path=remote_path)
    logging.info("File is being uploaded...")
    logging.info("You can use 'inductiva storage ls' to check the status.")


def _convert_path_unix(path):
    """
    Convert a Windows path to a Unix-style path.
    """
    _, path = os.path.splitdrive(path)
    if "\\" in path:
        path = str(pathlib.PureWindowsPath(path).as_posix())
    return path


def _construct_remote_paths(local_path, remote_dir):
    """ Constructs remote paths for files to be uploaded."""
    remote_dir = os.path.normpath(remote_dir)
    local_path = os.path.normpath(_convert_path_unix(local_path))

    is_dir = os.path.isdir(local_path)
    if is_dir:
        local_dir = os.path.normpath(local_path)
        file_paths, total_size = _list_files(local_path)

        remote_file_paths = [
            os.path.normpath(os.path.join(remote_dir, file_path))
            for file_path in file_paths
        ]
    else:
        local_dir = os.path.dirname(local_path)
        filename = os.path.basename(local_path)
        remote_file_paths = [os.path.join(remote_dir, filename)]
        total_size = os.path.getsize(local_path)

    remote_file_paths = [
        _convert_path_unix(remote_path) for remote_path in remote_file_paths
    ]
    return remote_file_paths, local_dir, total_size


def upload(
    local_path: str,
    remote_dir: str,
):
    """
    Upload a local file or directory to a specified remote directory.

    Args:
        local_path (str): The path to the local file or directory to be
            uploaded.
        remote_dir (str): The remote directory where the file will
            be uploaded.

    Example:
        Upload a file to a remote directory:

        .. code-block:: python

            inductiva.storage.upload('local/path/file.txt', 'my_data')

        Upload a directory to a remote location:

        .. code-block:: python

            inductiva.storage.upload('local/path/folder', 'my_data')
    """

    if not os.path.exists(local_path):
        raise ValueError(f"File or directory '{local_path}' does not exist.")

    remote_file_paths, local_dir, total_size = _construct_remote_paths(
        local_path, remote_dir)

    if os.path.join(remote_dir, constants.TASK_OUTPUT_ZIP) in remote_file_paths:
        raise ValueError(f"Invalid file name: '{constants.TASK_OUTPUT_ZIP}.'")

    logging.info("Uploading content...")

    api_instance = inductiva.client.StorageApi(inductiva.api.get_client())

    urls = get_signed_urls(paths=remote_file_paths, operation="upload")

    with tqdm.tqdm(total=total_size,
                   unit="B",
                   unit_scale=True,
                   unit_divisor=1000) as progress_bar:

        for url, remote_file_path in zip(urls, remote_file_paths):
            file_path = remote_file_path.removeprefix(f"{remote_dir}/")
            local_file_path = _convert_path_unix(
                os.path.join(local_dir, file_path))

            try:
                methods.upload_file(api_instance, local_file_path, "PUT", url,
                                    progress_bar)
            except exceptions.ApiException as e:
                raise e

    logging.info("Content uploaded successfully.")


def _resolve_local_path(
    url,
    remote_base_path,
    local_base_dir,
    append_path=None,
    strip_zip=False,
):
    remote_url_path = urllib.parse.urlparse(url).path
    index = remote_url_path.find(remote_base_path)
    relative_path = remote_url_path[index:]

    if strip_zip and relative_path.endswith(".zip"):
        relative_path = relative_path.removesuffix(".zip")

    local_path = os.path.join(local_base_dir, relative_path)
    if append_path:
        local_path = os.path.join(local_path, append_path)

    os.makedirs(os.path.dirname(local_path), exist_ok=True)

    return local_path


def _get_size(url, pool_manager):
    response = pool_manager.urlopen("HEAD", url)
    size = int(response.getheader("Content-Length"))
    response.release_conn()
    return size


def _download_file(url,
                   download_path,
                   progress_bar,
                   progress_bar_lock,
                   pool_manager,
                   range_start=None,
                   range_end=None):
    headers = {}
    is_range = range_start is not None and range_end is not None

    if is_range:
        headers["Range"] = f"bytes={range_start}-{range_end}"
    response = pool_manager.urlopen("GET",
                                    url,
                                    headers=headers,
                                    preload_content=False)

    with open(download_path, "wb") as file:
        for chunk in response.stream():
            file.write(chunk)
            with progress_bar_lock:
                progress_bar.update(len(chunk))
    response.release_conn()

    return download_path


def _is_file_inside_zip(path):
    parts = path.split(os.sep)
    return any(part.endswith(".zip") for part in parts[:-1])


def _get_progress_bar(desc, total_bytes):
    return tqdm.tqdm(
        total=total_bytes,
        unit="B",
        unit_scale=True,
        unit_divisor=1000,
        desc=desc,
    )


def _download_file_from_inside_zip(remote_path, local_dir, pool_manager):
    before, after = remote_path.split(".zip" + os.sep, 1)
    path = before + ".zip"
    zip_relative_path = os.path.dirname(after)
    zip_filename = os.path.basename(after)

    url = get_signed_urls(paths=[path], operation="download")[0]

    if constants.TASK_OUTPUT_ZIP in path:
        prefix = data.ARTIFACTS_DIRNAME
    elif constants.TASK_INPUT_ZIP in path:
        prefix = data.INPUT_DIRNAME
    else:
        prefix = ""

    zip_files = get_zip_contents(path, prefix + zip_relative_path).files

    for zip_file in zip_files:
        if zip_file.name == zip_filename:
            break
    else:
        raise ValueError(f"File \"{after}\" not found in \"{path}\".")

    byte_range = get_file_range(path, prefix + zip_relative_path, zip_filename)
    range_start = byte_range.range_start
    range_end = byte_range.range_end
    compress_type = zip_file.compress_type
    file_path = after + ".zip" if compress_type else after
    download_path = _resolve_local_path(url, path, local_dir, file_path, True)

    desc = f"Downloading \"{zip_filename}\" from \"{remote_path}\""
    with _get_progress_bar(desc, range_start - range_end) as progress_bar:
        progress_bar_lock = threading.Lock()
        return _download_file(
            url=url,
            download_path=download_path,
            progress_bar=progress_bar,
            progress_bar_lock=progress_bar_lock,
            pool_manager=pool_manager,
            range_start=range_start,
            range_end=range_end,
        )


def _download_path(remote_path, local_dir, pool_manager):
    urls = get_signed_urls(paths=[remote_path], operation="download")
    with concurrent.futures.ThreadPoolExecutor() as executor:
        total_bytes = sum(
            executor.map(_get_size, urls, itertools.repeat(pool_manager)))

    resolved_paths = [
        _resolve_local_path(url, remote_path, local_dir) for url in urls
    ]

    desc = f"Downloading {len(urls)} file(s) from \"{remote_path}\""
    with _get_progress_bar(desc, total_bytes) as progress_bar:
        with concurrent.futures.ThreadPoolExecutor() as executor:
            progress_bar_lock = threading.Lock()
            return list(
                executor.map(_download_file, urls, resolved_paths,
                             itertools.repeat(progress_bar),
                             itertools.repeat(progress_bar_lock),
                             itertools.repeat(pool_manager)))


def _decompress(paths):
    for path in paths:
        decompress_dir, ext = os.path.splitext(path)
        # TODO: Improve the check for ZIP file
        if ext != ".zip":
            return
        utils.data.decompress_zip(path, decompress_dir)
        os.remove(path)


def _decompress_file_inside_zip(path):
    if not path.endswith(".zip"):
        return path

    with open(path, "rb") as f:
        compressed_data = f.read()
    decompressed_data = zlib.decompress(compressed_data, -zlib.MAX_WBITS)

    decompress_path = path.removesuffix(".zip")
    with open(decompress_path, "wb") as f:
        f.write(decompressed_data)
    os.remove(path)

    return decompress_path


def download(remote_path: str, local_dir: str = "", decompress: bool = True):
    """
    Downloads a file or folder from storage to a local directory, optionally
    decompressing the contents.

    Args:
        remote_path (str): The path of the file or folder on the remote server
            to download.
        local_dir (str, optional): The local directory where the file or folder
            will be saved. Defaults to the current working directory.
        decompress (bool, optional): Whether to decompress the downloaded file
            or folder if it is compressed. Defaults to True.

    Examples:
        Download a folder from a remote server to the current directory:

        .. code-block:: python

            inductiva.storage.download(remote_path="/path/to/remote/folder/")

        Download a file and save it to a local directory without decompressing:

        .. code-block:: python

            inductiva.storage.download(remote_path="/path/to/remote/file.zip",
                                       local_dir="/local/directory",
                                       decompress=False)

        Download a file inside a zip archive:

        .. code-block:: python

            inductiva.storage.download(
                remote_path="/some_task_id/output.zip/stdout.txt"
            )
    Note:
        It is not possible to download folders that are inside zip archives.
    """

    api_instance = inductiva.client.StorageApi(inductiva.api.get_client())
    pool_manager = api_instance.api_client.rest_client.pool_manager

    if _is_file_inside_zip(remote_path):
        download_path = _download_file_from_inside_zip(remote_path, local_dir,
                                                       pool_manager)
        paths = [_decompress_file_inside_zip(download_path)]
    else:
        paths = _download_path(remote_path, local_dir, pool_manager)

    if decompress:
        _decompress(paths)

    download_path = (paths[0] if len(paths) == 1 else os.path.join(
        local_dir, remote_path))

    logging.info('Successfully downloaded %d file(s) to "%s".', len(paths),
                 download_path)


def _list_files(root_path: str) -> Tuple[List[str], int]:
    """
    Lists all files within a directory and returns their total size.

    Args:
        root_path: The root directory to list files from.
    """
    file_paths = []
    total_size = 0

    for dirpath, _, filenames in os.walk(root_path):
        for filename in filenames:
            path = os.path.relpath(os.path.join(dirpath, filename), root_path)
            file_paths.append(path)

            full_path = os.path.join(dirpath, filename)
            total_size += os.path.getsize(full_path)

    return file_paths, total_size


def remove(remote_path: str):
    """
    Removes a file or directory from the remote location.

    Parameters:
    - remote_path (str): The path to the remote file or directory.
    """
    logging.info("Removing '%s' from remote storage...", remote_path)

    # Since we don't allow root files in workspaces it must be a directory
    # otherwise path validation in the backend will give error
    if "/" not in remote_path:
        remote_path += "/"

    api = inductiva.client.StorageApi(inductiva.api.get_client())
    api.delete_file(path=remote_path)

    logging.info("Successfully removed '%s' from remote storage.", remote_path)


def copy(source: str, target: str):
    """
    Copies a file or folder from a source path in storage to a target path.

    Args:
        source (str): The source path of the file or directory to copy.
        target (str): The destination path where the file or directory
                      should be copied to.
    """
    api = inductiva.client.StorageApi(inductiva.api.get_client())
    api.copy(source=source, target=target)
    logging.info("Copied %s to %s successfully.", source, target)


class StorageOperation():
    """Represents a storage operation running remotely via Inductiva API."""

    def __init__(self, api: inductiva.client.StorageApi, id_):
        self._api = api
        self.id = id_

    def _update_from_api_response(self, response: models.StorageOperation):
        self._name = response.name
        self._status = response.status
        self._attributes = response.attributes
        self._start_time = response.start_time
        self._end_time = response.end_time
        self._error_message = response.error_message

    @classmethod
    def from_api_response(cls, api, response):
        op = cls(api, response["id"])

        op._update_from_api_response(response,)
        return op

    def _refresh(self):
        resp = self._api.get_operation(operation_id=self.id)

        self._update_from_api_response(resp)

    def wait(self, poll_s: int = 2):
        """Wait for the operation to complete.

        Args:
            poll_s: Time in seconds between calls to the API to update
                the status of the operation.
        """

        while self._status == models.OperationStatus.RUNNING:
            self._refresh()
            time.sleep(poll_s)

        if self._status == models.OperationStatus.FAILED:
            logging.error("Operation failed: %s", self._error_message)
        else:
            logging.info("Operation completed successfully.")

        return self._status


class ExportDestination(Enum):
    AWS_S3 = "aws-s3"

    def __str__(self):
        return self.value


def _initiate_multipart_upload(filename, bucket_name, region_name):
    """
    Initiate a multipart upload on S3 and return the UploadId.
    """
    s3_client = boto3.client("s3",
                             region_name=region_name,
                             config=Config(region_name=region_name,
                                           signature_version="v4"))
    response = s3_client.create_multipart_upload(
        Bucket=bucket_name,
        Key=filename,
    )
    return response["UploadId"]


def _generate_presigned_url(
    upload_id,
    part_number,
    filename,
    bucket_name,
    region_name,
):
    """
    Generate a presigned URL for uploading a part to S3.
    """
    s3_client = boto3.client("s3",
                             region_name=region_name,
                             config=Config(region_name=region_name,
                                           signature_version="v4"))
    method_parameters = {
        "Bucket": bucket_name,
        "Key": filename,
        "PartNumber": part_number,
        "UploadId": upload_id,
    }

    signed_url = s3_client.generate_presigned_url(
        "upload_part",
        Params=method_parameters,
        ExpiresIn=3600,
    )

    return signed_url


def _generate_complete_multipart_upload_signed_url(
    upload_id,
    filename,
    bucket_name,
    region_name,
):
    """
    Generate a presigned URL for completing the multipart upload.
    """
    s3_client = boto3.client("s3",
                             region_name=region_name,
                             config=Config(region_name=region_name,
                                           signature_version="v4"))

    signed_url = s3_client.generate_presigned_url(
        ClientMethod="complete_multipart_upload",
        Params={
            "Bucket": bucket_name,
            "Key": filename,
            "UploadId": upload_id
        },
        HttpMethod="POST",
    )

    return signed_url


def _get_file_size(file_path):
    api = inductiva.client.StorageApi(inductiva.api.get_client())

    contents = api.list_storage_contents(path=file_path, max_results=2)
    if len(contents) > 1:
        raise ValueError(f"Multiple files found at {file_path}. "
                         "Please specify a single file.")

    return list(contents.values())[0].size_bytes


def _get_multipart_parts(size: int,
                         part_size: int = 128 * MB) -> Tuple[int, int]:
    """
    Calculate the size of each part and the total number of parts

    The goal is to divide the data into parts `part_size` each:
    1. No more than 10,000 parts are created (maximum parts allowed by S3).
    2. The part size might be increased to avoid exceeding the part limit.
    3. The part size cannot be lower than 5MB.


    Args:
        size: The total size of the file to be uploaded, in bytes.
        part_size: The minimum size of each part, in bytes.

    Returns:
        - part_size: The size of each part in bytes.
        - part_count: The total number of parts.
    """
    max_parts = 10000
    min_allowed_part_size = 5 * MB  # 5MB

    # Ensure part_size is at least 5MB
    part_size = max(part_size, min_allowed_part_size)

    if size <= part_size:
        return size, 1

    # Calculate the part size based on the smaller of two values:
    # - At least `part_size`
    # - Maximum size to ensure no more than 10,000 parts (size // 10000)
    max_allowed_part_size = size // max_parts
    part_size = max(part_size, max_allowed_part_size)

    part_count = math.ceil(size / part_size)

    return part_size, part_count


def multipart_upload(
    path,
    parts_size,
    upload_parts,
    complete_multipart_url,
):
    """
    Perform the multipart upload using the server.
    """
    api = inductiva.client.StorageApi(inductiva.api.get_client())

    api.export_multipart_files(
        multi_part_export_operation=models.MultiPartExportOperation(
            path=path,
            parts_size=parts_size,
            upload_parts=upload_parts,
            complete_multipart_url=complete_multipart_url,
        ))


def export_to_aws_s3(path_to_export, part_size, filename, bucket_name):
    if not _boto3_imported:
        print("boto3 is not installed. Please run "
              "'pip install inductiva[aws]' to install it.")
        return
    try:
        boto3.client("sts").get_caller_identity()
    except Exception:  # pylint: disable=broad-exception-caught
        print("AWS credentials not found. Please set your "
              "AWS credentials with 'aws configure'.")
        return
    try:
        boto3.client("s3").head_bucket(Bucket=bucket_name)
    except Exception:  # pylint: disable=broad-exception-caught
        print(f"Bucket {bucket_name} not found. Make sure the bucket exists "
              "and you have the correct permissions: "
              "https://inductiva.ai/guides/how-it-works/recipes"
              "/export-files-to-aws/index")
        return

    region_name = boto3.Session().region_name
    if not region_name:
        print("AWS region not found. Please set your AWS region with "
              "'aws configure'.")
        return

    # Step 1: Get the file size
    file_size = _get_file_size(path_to_export)

    # Step 2: Calculate the part size and count
    parts_size, parts_count = _get_multipart_parts(
        file_size,
        part_size=part_size * MB,
    )

    # Step 3: Initiate the multipart upload on aws
    upload_id = _initiate_multipart_upload(filename, bucket_name, region_name)

    # Step 4: Generate presigned URLs for each part
    upload_parts = []
    for part_number in range(1, parts_count + 1):
        presigned_url = _generate_presigned_url(upload_id, part_number,
                                                filename, bucket_name,
                                                region_name)
        upload_parts.append({
            "part_number": part_number,
            "part_url": presigned_url
        })

    # Step 5: Generate the complete multipart upload signed URL
    complete_multipart_url = _generate_complete_multipart_upload_signed_url(
        upload_id, filename, bucket_name, region_name)

    # Step 6: Ask the server to perform the multipart upload
    multipart_upload(
        path_to_export,
        parts_size,
        upload_parts,
        complete_multipart_url,
    )
    print(
        "Export is being done by inductiva server. You can close the terminal.")


def export(
    path_to_export: str,
    export_to: ExportDestination,
    bucket_name: str,
    file_name: Optional[str] = None,
    part_size: int = 128,
):
    file_name = file_name or pathlib.Path(path_to_export).name
    if export_to == ExportDestination.AWS_S3:
        print(f"Exporting {path_to_export} to {bucket_name}...")
        export_to_aws_s3(
            path_to_export,
            part_size,
            file_name,
            bucket_name,
        )
    else:
        raise ValueError(f"Unsupported export destination: {export_to}")



# coding: utf-8

"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 0.1.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from datetime import datetime
from pydantic import BaseModel, ConfigDict, StrictBool, StrictFloat, StrictInt, StrictStr
from typing import Any, ClassVar, Dict, List, Optional, Union
from inductiva.client.models.machine_group_costs import MachineGroupCosts
from inductiva.client.models.machine_group_status import MachineGroupStatus
from inductiva.client.models.machine_group_type import MachineGroupType
from inductiva.client.models.machine_group_usage_statistics import MachineGroupUsageStatistics
from inductiva.client.models.providers import Providers
from typing import Optional, Set
from typing_extensions import Self


class MachineGroupWithCosts(BaseModel):
    """
    MachineGroupWithCosts
    """

  # noqa: E501
    id: StrictStr
    name: StrictStr
    machine_type: StrictStr
    disk_size_gb: StrictInt
    auto_resize_disk_max_gb: Optional[StrictInt] = None
    num_vms: StrictInt
    min_vms: Optional[StrictInt] = None
    max_vms: Optional[StrictInt] = None
    spot: StrictBool
    type: MachineGroupType
    creation_timestamp: datetime
    start_timestamp: Optional[datetime]
    deletion_timestamp: Optional[datetime]
    provider_id: Providers
    provider_description: Optional[StrictStr] = None
    status: MachineGroupStatus
    threads_per_core: StrictInt
    max_idle_time: Optional[Union[StrictFloat, StrictInt]]
    termination_reason: Optional[StrictStr] = None
    statistics: MachineGroupUsageStatistics
    idle_time_seconds: Optional[StrictInt] = None
    costs: MachineGroupCosts
    __properties: ClassVar[List[str]] = [
        "id", "name", "machine_type", "disk_size_gb", "auto_resize_disk_max_gb",
        "num_vms", "min_vms", "max_vms", "spot", "type", "creation_timestamp",
        "start_timestamp", "deletion_timestamp", "provider_id",
        "provider_description", "status", "threads_per_core", "max_idle_time",
        "termination_reason", "statistics", "idle_time_seconds", "costs"
    ]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of MachineGroupWithCosts from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of statistics
        if self.statistics:
            _dict['statistics'] = self.statistics.to_dict()
        # override the default output from pydantic by calling `to_dict()` of costs
        if self.costs:
            _dict['costs'] = self.costs.to_dict()
        # set to None if auto_resize_disk_max_gb (nullable) is None
        # and model_fields_set contains the field
        if self.auto_resize_disk_max_gb is None and "auto_resize_disk_max_gb" in self.model_fields_set:
            _dict['auto_resize_disk_max_gb'] = None

        # set to None if min_vms (nullable) is None
        # and model_fields_set contains the field
        if self.min_vms is None and "min_vms" in self.model_fields_set:
            _dict['min_vms'] = None

        # set to None if max_vms (nullable) is None
        # and model_fields_set contains the field
        if self.max_vms is None and "max_vms" in self.model_fields_set:
            _dict['max_vms'] = None

        # set to None if start_timestamp (nullable) is None
        # and model_fields_set contains the field
        if self.start_timestamp is None and "start_timestamp" in self.model_fields_set:
            _dict['start_timestamp'] = None

        # set to None if deletion_timestamp (nullable) is None
        # and model_fields_set contains the field
        if self.deletion_timestamp is None and "deletion_timestamp" in self.model_fields_set:
            _dict['deletion_timestamp'] = None

        # set to None if provider_description (nullable) is None
        # and model_fields_set contains the field
        if self.provider_description is None and "provider_description" in self.model_fields_set:
            _dict['provider_description'] = None

        # set to None if max_idle_time (nullable) is None
        # and model_fields_set contains the field
        if self.max_idle_time is None and "max_idle_time" in self.model_fields_set:
            _dict['max_idle_time'] = None

        # set to None if termination_reason (nullable) is None
        # and model_fields_set contains the field
        if self.termination_reason is None and "termination_reason" in self.model_fields_set:
            _dict['termination_reason'] = None

        # set to None if idle_time_seconds (nullable) is None
        # and model_fields_set contains the field
        if self.idle_time_seconds is None and "idle_time_seconds" in self.model_fields_set:
            _dict['idle_time_seconds'] = None

        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of MachineGroupWithCosts from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "id":
                obj.get("id"),
            "name":
                obj.get("name"),
            "machine_type":
                obj.get("machine_type"),
            "disk_size_gb":
                obj.get("disk_size_gb"),
            "auto_resize_disk_max_gb":
                obj.get("auto_resize_disk_max_gb"),
            "num_vms":
                obj.get("num_vms"),
            "min_vms":
                obj.get("min_vms"),
            "max_vms":
                obj.get("max_vms"),
            "spot":
                obj.get("spot"),
            "type":
                obj.get("type"),
            "creation_timestamp":
                obj.get("creation_timestamp"),
            "start_timestamp":
                obj.get("start_timestamp"),
            "deletion_timestamp":
                obj.get("deletion_timestamp"),
            "provider_id":
                obj.get("provider_id"),
            "provider_description":
                obj.get("provider_description"),
            "status":
                obj.get("status"),
            "threads_per_core":
                obj.get("threads_per_core"),
            "max_idle_time":
                obj.get("max_idle_time"),
            "termination_reason":
                obj.get("termination_reason"),
            "statistics":
                MachineGroupUsageStatistics.from_dict(obj["statistics"])
                if obj.get("statistics") is not None else None,
            "idle_time_seconds":
                obj.get("idle_time_seconds"),
            "costs":
                MachineGroupCosts.from_dict(obj["costs"])
                if obj.get("costs") is not None else None
        })
        return _obj



"""
This module provides functionality for managing projects and their associated
tasks within the Inductiva platform. A project serves as a container for
grouping related tasks, enabling better organization and management of
computational workflows.

Classes:
    - Project: Represents a project that groups related tasks together.

Functions:
    - get_projects(): Retrieves all projects associated with the current user.

Key Features:
    - Create and manage projects on the backend.
    - Add tasks to projects and retrieve the most recent tasks or filter by
      status.
    - Monitor the status of all tasks in a project and wait for their
    completion.
    - Download outputs for all tasks in a project.
    - Estimate the total computation cost of a project based on its tasks.

Example Usage:

      .. code-block:: python

        import inductiva

        # Create a new project or load an existing one
        project = inductiva.projects.Project("my_project")

        # Select a simulator
        fvcom = inductiva.simulators.FVCOM()
        # Run and add tasks to the project
        for i in range(10):
            task = fvcom.run(...)
            project.add_task(task)

        # Monitor task completion
        project.wait()
        # Download outputs for all tasks
        project.download_outputs()
        # Print project details
        print(project)
"""

import logging
import time
from typing import List, Optional

from inductiva import tasks
from inductiva import api as inductiva_api
import inductiva.client
from inductiva.client.models import TaskStatusCode, ProjectCreate, ProjectType
from inductiva.client import ApiException

# from inductiva.client import models
from inductiva.utils import files, format_utils

_logger = logging.getLogger(__name__)


def get_projects() -> List["Project"]:
    """Gets all the user's projects."""
    try:
        _logger.debug("Trying to get remote projects")
        api = inductiva.client.ProjectsApi(inductiva_api.get_client())
        response = api.get_user_projects()
    except ApiException as ex:
        _logger.error("Failed to get remote projects", exc_info=ex)
        raise ex

    # pylint: disable=protected-access
    return [Project._from_api_response(resp) for resp in response]


class Project:
    """
    Projects management class.

    Groups related tasks together under a single project.

    Example:

      .. code-block:: python

        project = inductiva.projects.Project("my project")

        task_1 = simulator.run(...)
        project.add_task(task_1)

        task_2 = simulator.run(...)
        project.add_task(task_2)
    """

    def __init__(self, name: str):
        """
        Initialize the Project instance.

        Args:
          name (str): The name of the project.
        """
        self._api = inductiva.client.ProjectsApi(inductiva_api.get_client())
        # If the project already exists, we will load it from the backend.
        self._proj_data = self._get_project(name)
        # Else, we will create a new project.
        if not self._proj_data:
            self._proj_data = self._create_project(name)

    def _get_project_type(self):
        return ProjectType.PROJECT

    def _get_project(self, name: str):
        """Fetches the project info from the backend."""
        try:
            return self._api.get_project(name=name)
        except ApiException as ex:
            if ex.status != 404:
                _logger.error("Failed to get project %s", name, exc_info=ex)
                raise ex
        return None

    def _create_project(self, name):
        """Creates a project with the given name on the backend."""
        try:
            return self._api.create_project(project_create=ProjectCreate(
                name=name, project_type=self._get_project_type()))
        except ApiException as ex:
            _logger.error("Failed to create project %s", name, exc_info=ex)
            raise RuntimeError(f"Unable to create project {name}") from ex

    @classmethod
    def _from_api_response(cls, resp):
        """Creates a Project instance from the API response."""
        project = cls.__new__(cls)
        project._proj_data = resp
        return project

    @property
    def name(self) -> str:
        """Returns the name of the project."""
        return self._proj_data.name

    @property
    def created_at(self) -> str:
        """Returns the creation date and time of the project."""
        return self._proj_data.created_at.isoformat()

    @property
    def num_tasks(self) -> int:
        """Returns the number of tasks in the project."""
        return self._proj_data.num_tasks

    @property
    def id(self) -> str:
        """Returns the unique ID of the project."""
        return self._proj_data.id

    @property
    def task_by_status(self) -> dict:
        """
        Returns a dictionary with the number of tasks by status.
        The keys are the status codes and the values are the number of tasks
        with that status.
        """
        return {
            TaskStatusCode(attr): int(value)
            for attr, value in self._proj_data.task_status_overview.items()
        }

    @property
    def estimated_computation_cost(self) -> float:
        """
        Returns the estimated project cost.

        Computed as the sum of the estimated computation cost of each task.
        """
        return self._proj_data.estimated_computation_cost

    def __str__(self) -> str:
        formatted_cost = format_utils.currency_formatter(
            self.estimated_computation_cost)
        formatted_created_at = format_utils.datetime_formatter_ymd_hm(
            self.created_at)
        status_report = "\n".join(
            f"  {k}: {v}" for k, v in self.task_by_status.items())

        return (f"Project '{self.name}' created at {formatted_created_at}.\n"
                f"\nTotal number of tasks: {self.num_tasks}\n"
                "\nTasks by status:\n"
                f"{status_report}\n"
                f"\nEstimated total computation cost: {formatted_cost}\n")

    def add_task(self, task: tasks.Task):
        """
        Adds a task to the project.

        Args:
            task: The task to add to the project.
        """
        try:
            self._api.add_task_to_project(name=self.name, task_id=task.id)
        except ApiException as ex:
            _logger.error(
                "Failed to add task %s to project %s",
                task.id,
                self.name,
                exc_info=ex,
            )
            raise ex

    def get_tasks(self,
                  last_n: int = -1,
                  status: Optional[str] = None) -> List[tasks.Task]:
        """
        Get the the tasks of this project.

        Optionally, those can be filtered by task status.

        Args:
            last_n (int): The number of tasks with repect to the submission
                time to fetch. If `last_n<=0` we fetch all tasks submitted
                to the project.
            status: Status of the tasks to get. If `None`, tasks with any
                status will be returned.
        """
        return tasks.get_tasks(last_n=last_n, project=self.name, status=status)

    def wait(self):
        """Wait for all the tasks in a project to complete."""
        all_tasks = self.get_tasks()
        print("Waiting for ALL tasks to finish")
        while not all(x.is_terminal() for x in all_tasks):
            finished = sum(x.is_terminal() for x in all_tasks)
            print(f"Finished: {finished} Total: {len(all_tasks)}", end="\r")
            time.sleep(5)
        print("All tasks in the project terminated.")

    def download_outputs(self, output_dir: Optional[str] = None):
        """Downloads all the outputs for all the tasks in the project.

        All task outputs will be organized within the specified `output_dir`.
        If `output_dir` is not provided, outputs will be saved to a default
        location under `inductiva_output/<project_name>/<task_id>/`.
        Otherwise, they will be stored in `<output_dir>/<task_id>/`.

        Args:
            output_dir (str, optional): The base directory where project outputs
                                        will be downloaded.
        """

        for task in self.get_tasks():
            base_path = output_dir or files.resolve_output_path(self.name)
            task.download_outputs(output_dir=f"{base_path}/{task.id}")

    def __eq__(self, other) -> bool:
        return (isinstance(other, Project) and self.name == other.name and
                self.id == other.id)



# coding: utf-8

# flake8: noqa

"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 0.1.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501

__version__ = "1.0.0"

# import apis into sdk package
from inductiva.client.api.compute_api import ComputeApi
from inductiva.client.api.events_api import EventsApi
from inductiva.client.api.projects_api import ProjectsApi
from inductiva.client.api.simulators_api import SimulatorsApi
from inductiva.client.api.storage_api import StorageApi
from inductiva.client.api.tasks_api import TasksApi
from inductiva.client.api.users_api import UsersApi
from inductiva.client.api.version_api import VersionApi

# import ApiClient
from inductiva.client.api_response import ApiResponse
from inductiva.client.api_client import ApiClient
from inductiva.client.configuration import Configuration
from inductiva.client.exceptions import OpenApiException
from inductiva.client.exceptions import ApiTypeError
from inductiva.client.exceptions import ApiValueError
from inductiva.client.exceptions import ApiKeyError
from inductiva.client.exceptions import ApiAttributeError
from inductiva.client.exceptions import ApiException

# import models into sdk package
from inductiva.client.models.action import Action
from inductiva.client.models.action1 import Action1
from inductiva.client.models.action_email_create import ActionEmailCreate
from inductiva.client.models.action_email_info import ActionEmailInfo
from inductiva.client.models.action_webhook_create import ActionWebhookCreate
from inductiva.client.models.action_webhook_info import ActionWebhookInfo
from inductiva.client.models.autoscale_policy import AutoscalePolicy
from inductiva.client.models.available_simulator_versions import AvailableSimulatorVersions
from inductiva.client.models.backend_version import BackendVersion
from inductiva.client.models.bucket import Bucket
from inductiva.client.models.bulk_simulator_operation import BulkSimulatorOperation
from inductiva.client.models.bulk_simulator_response import BulkSimulatorResponse
from inductiva.client.models.cpu_info import CPUInfo
from inductiva.client.models.capability import Capability
from inductiva.client.models.compression_method import CompressionMethod
from inductiva.client.models.cost_components import CostComponents
from inductiva.client.models.cost_detail import CostDetail
from inductiva.client.models.cost_type import CostType
from inductiva.client.models.currency_code import CurrencyCode
from inductiva.client.models.disk_resize_request import DiskResizeRequest
from inductiva.client.models.dynamic_disk_resize_config import DynamicDiskResizeConfig
from inductiva.client.models.event_create import EventCreate
from inductiva.client.models.event_info import EventInfo
from inductiva.client.models.executer import Executer
from inductiva.client.models.file_info import FileInfo
from inductiva.client.models.gpu_info import GPUInfo
from inductiva.client.models.http_validation_error import HTTPValidationError
from inductiva.client.models.machine_group_cost_per_hour import MachineGroupCostPerHour
from inductiva.client.models.machine_group_costs import MachineGroupCosts
from inductiva.client.models.machine_group_status import MachineGroupStatus
from inductiva.client.models.machine_group_time_stat import MachineGroupTimeStat
from inductiva.client.models.machine_group_type import MachineGroupType
from inductiva.client.models.machine_group_usage_statistics import MachineGroupUsageStatistics
from inductiva.client.models.machine_group_with_costs import MachineGroupWithCosts
from inductiva.client.models.machine_info import MachineInfo
from inductiva.client.models.machine_operation import MachineOperation
from inductiva.client.models.machine_operation_type import MachineOperationType
from inductiva.client.models.machine_type import MachineType
from inductiva.client.models.multi_part_export_operation import MultiPartExportOperation
from inductiva.client.models.operation_end import OperationEnd
from inductiva.client.models.operation_start import OperationStart
from inductiva.client.models.operation_status import OperationStatus
from inductiva.client.models.operation_type import OperationType
from inductiva.client.models.order import Order
from inductiva.client.models.organization_cost_components import OrganizationCostComponents
from inductiva.client.models.organization_costs import OrganizationCosts
from inductiva.client.models.processor_type import ProcessorType
from inductiva.client.models.project import Project
from inductiva.client.models.project_create import ProjectCreate
from inductiva.client.models.project_statistics import ProjectStatistics
from inductiva.client.models.project_type import ProjectType
from inductiva.client.models.project_update import ProjectUpdate
from inductiva.client.models.providers import Providers
from inductiva.client.models.quota import Quota
from inductiva.client.models.quota_scope import QuotaScope
from inductiva.client.models.register_vm_group_request import RegisterVMGroupRequest
from inductiva.client.models.select_mode import SelectMode
from inductiva.client.models.sharing_level import SharingLevel
from inductiva.client.models.simulator_create import SimulatorCreate
from inductiva.client.models.simulator_response import SimulatorResponse
from inductiva.client.models.simulator_update import SimulatorUpdate
from inductiva.client.models.simulator_version import SimulatorVersion
from inductiva.client.models.simulator_with_machine_types import SimulatorWithMachineTypes
from inductiva.client.models.spot import Spot
from inductiva.client.models.spot1 import Spot1
from inductiva.client.models.storage_contents_with_pagination import StorageContentsWithPagination
from inductiva.client.models.storage_cost import StorageCost
from inductiva.client.models.storage_file_info import StorageFileInfo
from inductiva.client.models.storage_operation import StorageOperation
from inductiva.client.models.storage_operation_name import StorageOperationName
from inductiva.client.models.storage_sort_by import StorageSortBy
from inductiva.client.models.task import Task
from inductiva.client.models.task_connection import TaskConnection
from inductiva.client.models.task_machine_operation import TaskMachineOperation
from inductiva.client.models.task_metric_create import TaskMetricCreate
from inductiva.client.models.task_metrics import TaskMetrics
from inductiva.client.models.task_operation import TaskOperation
from inductiva.client.models.task_operation_name import TaskOperationName
from inductiva.client.models.task_position_in_queue import TaskPositionInQueue
from inductiva.client.models.task_request import TaskRequest
from inductiva.client.models.task_runner_api_connection_info import TaskRunnerAPIConnectionInfo
from inductiva.client.models.task_runner_register_info import TaskRunnerRegisterInfo
from inductiva.client.models.task_status import TaskStatus
from inductiva.client.models.task_status_code import TaskStatusCode
from inductiva.client.models.task_status_info import TaskStatusInfo
from inductiva.client.models.task_step import TaskStep
from inductiva.client.models.task_submitted_info import TaskSubmittedInfo
from inductiva.client.models.task_with_pagination import TaskWithPagination
from inductiva.client.models.task_with_status_history import TaskWithStatusHistory
from inductiva.client.models.terms_and_conditions import TermsAndConditions
from inductiva.client.models.top_up_type import TopUpType
from inductiva.client.models.transaction import Transaction
from inductiva.client.models.trigger import Trigger
from inductiva.client.models.trigger1 import Trigger1
from inductiva.client.models.trigger_credits_info import TriggerCreditsInfo
from inductiva.client.models.trigger_credits_type import TriggerCreditsType
from inductiva.client.models.trigger_file_exists_observer_info import TriggerFileExistsObserverInfo
from inductiva.client.models.trigger_machine_group_create import TriggerMachineGroupCreate
from inductiva.client.models.trigger_machine_group_info import TriggerMachineGroupInfo
from inductiva.client.models.trigger_machine_group_type import TriggerMachineGroupType
from inductiva.client.models.trigger_observer_create import TriggerObserverCreate
from inductiva.client.models.trigger_observer_type import TriggerObserverType
from inductiva.client.models.trigger_regex_observer_info import TriggerRegexObserverInfo
from inductiva.client.models.trigger_task_create import TriggerTaskCreate
from inductiva.client.models.trigger_task_info import TriggerTaskInfo
from inductiva.client.models.trigger_task_type import TriggerTaskType
from inductiva.client.models.trigger_type import TriggerType
from inductiva.client.models.update_operation_status import UpdateOperationStatus
from inductiva.client.models.upload_part import UploadPart
from inductiva.client.models.usage_statistics import UsageStatistics
from inductiva.client.models.user import User
from inductiva.client.models.user_activity import UserActivity
from inductiva.client.models.user_computation_trend import UserComputationTrend
from inductiva.client.models.user_costs import UserCosts
from inductiva.client.models.user_costs_details import UserCostsDetails
from inductiva.client.models.user_costs_over_time import UserCostsOverTime
from inductiva.client.models.user_most_used_machine_types_overview import UserMostUsedMachineTypesOverview
from inductiva.client.models.user_most_used_simulators_overview import UserMostUsedSimulatorsOverview
from inductiva.client.models.user_task_status_overview import UserTaskStatusOverview
from inductiva.client.models.user_tasks_overview import UserTasksOverview
from inductiva.client.models.vm_group_config import VMGroupConfig
from inductiva.client.models.validation_error import ValidationError
from inductiva.client.models.validation_error_loc_inner import ValidationErrorLocInner
from inductiva.client.models.version_comparison_result import VersionComparisonResult
from inductiva.client.models.zip_archive_info import ZipArchiveInfo
from inductiva.client.models.zip_file_range import ZipFileRange



"""Loader module to register all modules within each package automatically."""
import importlib
import pkgutil
import logging

_logger = logging.getLogger(__name__)


def load_commands(parser,
                  module_dir,
                  package,
                  prefix="",
                  ignores_prefix=None,
                  hides_prefix=None):
    """Load all commands within the module_dir and package."""
    modules = pkgutil.iter_modules([module_dir])
    metavar = []
    for _, name, _ in modules:
        if not name.startswith(prefix) or \
            (ignores_prefix is not None and name.startswith(ignores_prefix)):
            continue

        modname = f".{name}"
        try:
            module = importlib.import_module(modname, package)
        except Exception as e:  # pylint: disable=broad-except
            _logger.error("Failed to import %s: %s", modname, e)
            continue

        if hasattr(module, "register"):
            indent = package.count(".") * "  "
            _logger.debug("%sRegistering %s", indent, modname)
            module.register(parser)
            metavar.append(name)

    if hasattr(parser, "choices") and hides_prefix:
        parser.metavar = (
            "{" +
            ",".join(cmd for cmd, name in zip(parser._name_parser_map, metavar)  # pylint: disable=protected-access
                     if not name.startswith(hides_prefix)) + "}")



"""Prints a task information via CLI."""
from typing import TextIO
import argparse
import sys

from inductiva import tasks, _cli


def task_info(args, fout: TextIO = sys.stdout):
    """Prints a task information."""
    task_id = args.id
    task = tasks.Task(task_id)
    info = task.get_info()
    print(info, file=fout)
    return 0


def register(parser):
    """Register the info tasks command."""
    subparser = parser.add_parser("info",
                                  help="Prints information related to a task.",
                                  formatter_class=argparse.RawTextHelpFormatter)

    subparser.description = ("The `inductiva tasks info` command provides "
                             "an extensive list of information about a task.")

    _cli.utils.add_watch_argument(subparser)
    subparser.add_argument("id",
                           type=str,
                           help="ID of the task to get information about.")
    subparser.set_defaults(func=task_info)



"""OpenSees module of the API."""
import logging
from typing import Literal, Optional, Union

from inductiva import simulators, tasks, types
from inductiva.commands.commands import Command
from inductiva.commands.mpiconfig import MPIConfig

AVAILABLE_OPENSEES_INTERFACES = ["python", "tcl", "eesd"]


@simulators.simulator.mpi_enabled
class OpenSees(simulators.Simulator):
    """Class to invoke a generic OpenSees simulation on the API."""

    def __init__(self,
                 /,
                 version: Optional[str] = None,
                 use_dev: bool = False,
                 interface: Literal["python", "tcl", "eesd"] = "python"):
        """Initialize the OpenSees simulator.

        Args:
            version (str): The version of the simulator to use. If None, the
                latest available version in the platform is used.
            use_dev (bool): Request use of the development version of
                the simulator. By default (False), the production version
                is used.
            interface (str): The interface to use for interacting with the
                simulator. Can be either "python" (default) or "tcl".
        """
        if interface.lower() not in AVAILABLE_OPENSEES_INTERFACES:
            raise ValueError(
                f"Interface '{interface}' for OpenSees is not supported. "
                f"Available interfaces are: "
                f"{AVAILABLE_OPENSEES_INTERFACES}")
        self._interface = interface.lower()
        super().__init__(version=version, use_dev=use_dev)
        self.simulator = "arbitrary_commands"
        self.simulator_name_alias = "opensees"

    @property
    def name(self):
        """Get the name of the simulator."""
        if self._interface == "python":
            return "OpenSeesPy"
        elif self._interface == "eesd":
            return "OpenSees-EESD"
        else:
            return "OpenSees"

    def _build_commands(self, sim_config_filename, mpi_config):
        """ Creates the list of commands based on the sim_config_filename and
        the provided mpi_config.
        """
        if self._interface == "python":
            return [
                Command(f"python {sim_config_filename}", mpi_config=mpi_config)
            ]

        if self._interface == "eesd":
            return [Command(f"OpenSees {sim_config_filename}")]

        return [
            Command(f"OpenSeesMP {sim_config_filename}", mpi_config=mpi_config)
        ]

    def _build_mpi_config(self, n_vcpus: int, use_hwthread: bool) -> MPIConfig:
        """ Creates an MPIConfig based on n_vcpus and use_hwthread.
        """
        return MPIConfig(version="4.1.6",
                         use_hwthread_cpus=use_hwthread,
                         **({
                             "np": n_vcpus
                         } if n_vcpus is not None else {}))

    def run(self,
            input_dir: Optional[str],
            *,
            on: types.ComputationalResources,
            sim_config_filename: Optional[str],
            n_vcpus: Optional[int] = None,
            use_hwthread: bool = True,
            storage_dir: Optional[str] = "",
            resubmit_on_preemption: bool = False,
            remote_assets: Optional[Union[str, list[str]]] = None,
            project: Optional[str] = None,
            time_to_live: Optional[str] = None,
            **kwargs) -> tasks.Task:
        """Run the simulation.

        Args:
            input_dir: Path to the directory of the simulation input files.
            on: The computational resource to launch the simulation on.
            sim_config_filename: Name of the simulation configuration file.
            n_vcpus: Number of vCPUs to use in the simulation. If not provided
                (default), all vCPUs will be used.
            use_hwthread: If specified Open MPI will attempt to discover the
                number of hardware threads on the node, and use that as the
                number of slots available.
            storage_dir: Directory for storing simulation results.
            resubmit_on_preemption (bool): Resubmit task for execution when
                previous execution attempts were preempted. Only applicable when
                using a preemptible resource, i.e., resource instantiated with
                `spot=True`.
            remote_assets: Additional remote files that will be copied to
                the simulation directory.
            project: Name of the project to which the task will be
                assigned. If None, the task will be assigned to
                the default project. If the project does not exist, it will be
                created.
            time_to_live: Maximum allowed runtime for the task, specified as a
                string duration. Supports common time duration formats such as
                "10m", "2 hours", "1h30m", or "90s". The task will be
                automatically terminated if it exceeds this duration after
                starting.
            other arguments: See the documentation of the base class.
        """

        if self._version == "2.5.0":
            logging.warning(
                "Opensees version 2.5.0 does not support parallel execution."
                "Changing to `n_vcpus` to 1.")
            n_vcpus = 1
            if self._interface == "python":
                raise ValueError(
                    "Opensees version 2.5.0 does not support `python` as"
                    " an interface. Please use `interface='tcl'`.")

        self._check_vcpus(n_vcpus, on)

        self._input_files_exist(input_dir=input_dir,
                                remote_assets=remote_assets,
                                sim_config_filename=sim_config_filename)

        mpi_config = self._build_mpi_config(n_vcpus, use_hwthread)

        if self._interface == "python":
            if n_vcpus or use_hwthread:
                logging.info("\nMPI flag detected (n_vcpus or use_hwthread)."
                             "\nThe simulation will run with mpirun "
                             "(Parallel Execution).")
                commands = self._build_commands(sim_config_filename, mpi_config)
            else:
                logging.info(
                    "\nNo MPI flag detected (n_vcpus or use_hwthread)."
                    "\nThe simulation will run sequentially with python.")
                commands = [f"python {sim_config_filename}"]
        else:
            commands = self._build_commands(sim_config_filename, mpi_config)

        return super().run(input_dir,
                           on=on,
                           commands=commands,
                           storage_dir=storage_dir,
                           resubmit_on_preemption=resubmit_on_preemption,
                           remote_assets=remote_assets,
                           project=project,
                           time_to_live=time_to_live,
                           **kwargs)



"""Upload storage contents via CLI."""

import argparse
from inductiva import storage


def upload(args):
    """Upload a file or folder to remote storage."""
    storage.upload(args.local_path, args.remote_dir)


def register(parser):
    """Register the upload command for the user's remote storage."""

    subparser = parser.add_parser(
        "upload",
        aliases=["ul"],
        help="Upload a file or folder to remote storage.",
        formatter_class=argparse.RawTextHelpFormatter)

    subparser.description = (
        "The `inductiva storage upload` command allows you to upload files "
        "or folders to your remote storage.\n"
        "Specify the local path and the remote destination.\n\n"
        "Example:\n"
        "    inductiva storage upload local/path/file_or_directory remote_dir\n"
    )

    subparser.add_argument(
        "local_path",
        type=str,
        help="The local path to the file or folder to upload.",
    )
    subparser.add_argument(
        "remote_dir",
        type=str,
        help=(
            "The remote directory where the uploaded content will be stored."),
    )

    subparser.set_defaults(func=upload)
